/*
 * ToDo: 
 * - Game Progression
 * 
 * Juice: 
 * - Ball transformieren
 */

import 'dart:html';
import 'dart:math';
import 'dart:web_gl';
import "dart:typed_data";

import 'package:asset_pack/asset_pack.dart';
import 'package:game_loop/game_loop_html.dart';

import 'webjuice.dart';
import "gens.dart";
import 'audiojuice.dart';

const double ROTATION_SPEED = 0.002;
const double CELL_SPACING = 0.1;
const double BALL_SIZE = 0.02;

const double THICKNESS = 0.1;

const double REBOUND_RAND = 0.0;

const double TAU = 6.283185307179586;

const double VOLUME_MUSIC = 0.7;
const double VOLUME_SCORE = 0.33;
const double VOLUME_WUSCH = 0.13;
const double VOLUME_SWELL = 0.2;
const double VOLUME_LOOSE = 0.8;

double sgn(x) {
  if(x < 0) {
    return -1.0;
  } else {
    return 1.0;
  }
}

bool _assetsDone = false;
bool _audioDone = false;

void error() {
  document.querySelector("#info").text = "Sorry. This game only works in a recent version of Google Chrome with WebGL enabled.";
}

void iOSLink() {
  final NodeValidatorBuilder _htmlValidator=new NodeValidatorBuilder.common()
    ..allowElement('a', attributes: ['href']);
  
  document.querySelector("#info").text = "";
  document.querySelector("#info").setInnerHtml("<a href=\"http://bit.ly/ballkeeper\">Get Ballkeeper for iOS in the App Store!</a>", validator: _htmlValidator);
}

void main() {
  
  var isChrome = window.navigator.appVersion.contains("Chrome");
  var iOS = window.navigator.appVersion.contains("iPad") || window.navigator.appVersion.contains("iPhone") || window.navigator.appVersion.contains("iPod");
  
  if(isChrome) {
    startLoading();
  } else if(iOS) {
    iOSLink();
  } else {
    error();
  }
}

void tryGo(gl, assets, audio) {
  if(_assetsDone && _audioDone) {
    go(gl, assets, audio);
  }
}

void go(gl, assets, audio) {
  HtmlElement headline = document.querySelector("#headline");
  HtmlElement info = document.querySelector("#info");
  
  headline.style.display = "none";
  info.style.display = "none";
  
  new Game(gl, assets, audio);  
}

void startLoading() {
  CanvasElement cnvs = document.querySelector("#cnvs");
  RenderingContext gl = cnvs.getContext("experimental-webgl");
  
  if(gl == null) {
    error();
    return;
  }
  
  AssetManager assets = new AssetManager();
  AudioManager audio = new AudioManager();
  
  audio.finishFunc = (() {
    _audioDone = true;
    tryGo(gl, assets, audio);
  });
  
  audio.load("collision", "assets/collision.ogg");
  audio.load("wusch", "assets/wusch.ogg");
  audio.load("wusch_long", "assets/wusch_long.ogg");
  audio.load("loose", "assets/loose.ogg");
  audio.load("music1", "assets/music1.ogg");
  audio.load("swell", "assets/swell.ogg");
  
  assets.loadPack("assets", "assets/assets.pack").then((pack) {
    _assetsDone = true;
    tryGo(gl, assets, audio);
  });  
}

class GameObject {
  void update(GameLoopHtml loop) {
    
  }
  
  void render(GameLoopHtml render) {
    
  }
}

abstract class OverRenders {
  bool finished();
  void render();
}

class FlashCircle implements OverRenders {
  Gen line;
  Gen radiusLine;
  Gen thicknessLine;
  
  bool finished() => line.finished();
  Game main;
  
  double time;
  double radius;
  double thickness;
    
  FlashCircle(this.main, this.radius, this.thickness, this.time) {
    line = new Line(1.0, 0.0, time);
    radiusLine = new Line(radius, 2.0, time);
    thicknessLine = new Line(thickness, thickness * 4.0, time);
  }
  
  void render() {
    GLProgram prog = main.programs["basic"];
    prog.bind();
    prog.uniforms["color"].set([1.0, 1.0, 1.0, line.value() * 0.15]);
    main.draw.arc(prog, 0, radiusLine.value(), thicknessLine.value(), 1.0 + (1/64));    
  }
}

class DarkCircle implements OverRenders {
  Gen line;
  Gen radiusLine;
  Gen thicknessLine;
  
  bool finished() => line.finished();
  Game main;
    
  DarkCircle(this.main) {
    double time = 4.0;
    line = new Line(0.0, 1.0, time);
    radiusLine = new Line(0.0, 3.0, time);
    thicknessLine = new Line(0.4, 0.4, time);
  }
  
  void render() {
    GLProgram prog = main.programs["basic"];
    prog.bind();
    prog.uniforms["color"].set([1.0, 1.0, 1.0, line.value() * 0.15]);
    main.draw.arc(prog, 0, radiusLine.value(), thicknessLine.value(), 1.0 + (1/64));    
  }
}




class Game {
  AssetPack assets;
  AssetManager assetManager;
  AudioManager audio;
  
  RenderingContext gl;
  GameLoopHtml loop;
  
  dynamic bgMusic;
  
  Map<String, GLProgram> programs;
  Map<String, GLTexture2D> textures;
  
  Random rng;
  
  List<CellLayer> objects;
  List<double> objectValues;
  
  List<OverRenders> overRenders;
  
  Ball ball;
  Background background;
  
  num rotation = 0.0;
  num targetRotation = 0.0;
  Gen rotationFilter;
  
  int level = 1;
  
  double deltaRotation;
  
  Cell lastCollisionCell;
  
  int points = 0;
    
  bool lost = false;
  bool started = false;
  
  Gen flashLine = new Const(0.0);
  
  GameDrawHelper draw;
  
  Game(this.gl, this.assetManager, this.audio) {
    assets = assetManager["assets"];
    
    programs = loadShaders(gl, assets["shaders"]);
    uploadTextures();
    
    buildRotationFilter();
    
        
    loop = new GameLoopHtml(gl.canvas);
    draw = new GameDrawHelper(this, programs);
    background = new Background(this);
    background.timeCanvas.redraw();
    
    rng = new Random();
    
    loop.onUpdate = (loop) => this.update(loop);
    loop.onRender = (loop) => this.render(loop);
    loop.onPointerLockChange = (loop) => this.enterLock(loop);
    loop.onKeyDown = (event) => this.keyDown(event);
    
    window.onResize.listen((e) => this.onResize());    
    
    setGenTime(loop.gameTime);
    
    
    onResize();
    
    background.timeCanvas.drawHelp();
    loop.start();
  }
  
  void keyDown(KeyboardEvent event) {
    if(started) {
      if(event.keyCode == 82) {
        start();
      }
    }
    if(event.keyCode == 70) {
      loop.enableFullscreen(!loop.isFullscreen);
      loop.pointerLock.requestLock();
    }
  }
  
  buildRotationFilter() {
    rotationFilter = new IIRFilter(new FuncGen(() => targetRotation), 0.13, 0.87);
  }
  
  void populate() {
    objectValues = [1/5, 2/5, 3/5, 4/5];
    objects = new List(objectValues.length);
    
    for(int i = 0; i < objectValues.length; i++) {
      var x = objectValues[i];
      var hitPoints = 1;
      if(i == objectValues.length - 1) {
        hitPoints = 2;
      }
      objects[i] = new CellLayer(this, 16, x, x*2.5, hitPoints);
    }
    
    ball = new Ball(this, 0, 0, 0.0);
    
    background.populate();
    overRenders = new List();
  }
  
  void repopulate() {
    flashLine = new Line(1.0, 0.0, 0.5);
    level++;
    ball.calculateSpeed();
    audio.play("wusch", VOLUME_WUSCH);
    
    for(int i = 0; i < objectValues.length; i++) {
      var obj = objects[i];
      if(obj == null || !obj.active) {
          if(i < (objectValues.length - 1)) {
            objects[i] = objects[i+1];
            objects[i].newRadius(objectValues[i]);
            objects[i].demote();
            objects[i+1] = null;
          } else {
            objects[i] = new CellLayer(this, 16, objectValues[i] , objectValues[i] * 3, 2);
          }
        }
      }
  }
  
  void score() {
    audio.play("collision", VOLUME_SCORE);
    background.jump = new Line(1.0, 0.0, 0.25);
    points += 1;
  }
  
  void start() {
    if(bgMusic != null) {
      bgMusic.stop(0);
    }
    
    level = 1;
    
    bgMusic = audio.play("music1", 0.5, true);
    audio.play("swell", VOLUME_SWELL);
    audio.play("wusch_long", VOLUME_WUSCH / 3);
    
    
    background.timeCanvas.redraw();
    points = 0;
    populate();
    started = true;
    lost = false;
    flashLine = new Line(1.0, 0.0, 2.0);
  }
  
  void uploadTextures() {
    Map<String, ImageElement> map = new Map();
    
    for(var key in assets.assets.keys) {
      if(assets[key] is ImageElement) {
        map[key] = assets[key];
      }
    }
    textures = loadTextures(gl, map, {});
  }
  
  void onResize() {
    gl.canvas.width = window.innerWidth;
    gl.canvas.height = window.innerHeight;
    draw.width = gl.canvas.width;
    draw.height = gl.canvas.height;
    draw.ratio = draw.height / draw.width;
    
    initGL();
    
    if(!started) {
      // update(loop);
      render(loop);
    }  
  }
  
  void loose() {
    if(!lost) {
      lost = true;
      if(bgMusic != null) {
        bgMusic.stop(0);
        bgMusic = null;
      }
      audio.play("loose", VOLUME_LOOSE);
      
      flashLine = new Line(1.0, 0.0, 1.0);
      background.alpha = new Line(1.0, 0.0, 2.0);
    }
  }
  
  void initGL() {
    gl.disable(DEPTH_TEST);
    gl.enable(BLEND);
    gl.viewport(0, 0, gl.canvas.width, gl.canvas.height);
    gl.clearColor(0.0, 0.0, 0.0, 1.0);
  }
  
  void update(GameLoopHtml loop) {
    if(lost && loop.mouse.pressed(0)) {
      start();
    }

    setGenTime(loop.gameTime);
    background.update(loop);
    
    if(!started) {
      return;
    }
    
    if(loop.pointerLock.locked) {
      var value = loop.mouse.dx * ROTATION_SPEED;
      
      var max = 0.135;
      
      if(value > max) {
        value = max;
      } else if(value < -max) {
        value = -max;
      }
      targetRotation = targetRotation - value;
    }
    
    var lastRotation = rotation;
    rotation = rotationFilter.value();
    deltaRotation = rotation - lastRotation;
    
    ball.update(loop);
    
    if((ball.x.abs() >= (objects.last.radius + 0.2) || ball.y.abs() >= (objects.last.radius + 0.2))
        && (ball.x.abs() > 1.3 || ball.y.abs() > 1.3)) {
      loose();
    }
    
    for(var object in objects) {
      object.update(loop);
    }
  }
  
  void render(GameLoopHtml loop) {
    gl.clear(COLOR_BUFFER_BIT);
    
    background.render(loop);
    if(started) {
      if(!lost) {
        for(var object in objects) {
          object.render(loop);
        }
        ball.render(loop);
      } else {
        draw.blendAdd();
        draw.color = [1.0, 1.0, 1.0, 1.0];
        draw.texQuadC(background.timeCanvas.texture, 0, 0, 1.3, 1.3);
      }
    }
    
    if(!started) {
      background.timeCanvas.drawHelp();
      draw.blendAdd();
      draw.color = [1.0, 1.0, 1.0, 1.0];
      draw.texQuadC(background.timeCanvas.texture, 0, 0.35, 1.0, 1.0);     
    }
    
    if(overRenders != null) {
      draw.blendAdd();
      overRenders.retainWhere((x) => !x.finished());
      for(var x in overRenders) {
        x.render();
      }
    }
    
    var flashValue = flashLine.value();
    if(flashValue > 0) {
      draw.blendAdd();
      draw.color = [1.0, 1.0 ,1.0, flashValue * 0.4];
      draw.texQuadC(textures["blur"], 0, 0, 3, 3);
    }
    
  }
  
  void enterLock(GameLoopHtml loop) {
    if(!started) {
      start();
    }
  }
}

class Ball implements GameObject {
  
  Game main;
  
  num x, y;
  num speed, phi;
  num size = BALL_SIZE;
  Gen sizeMul;
  
  double distanceToOrigin;
  double angleToOrigin;
  
  Ball(this.main, this.x, this.y, this.phi) {
    sizeMul = new Const(1.0);
    calculateSpeed();
  }
  
  void update(GameLoopHtml loop) {
    x = x + sin(phi) * speed * loop.dt;
    y = y + cos(phi) * speed * loop.dt;
    
    distanceToOrigin = sqrt(pow(x, 2) + pow(y, 2));
    angleToOrigin = (atan2(x, y) + TAU) % TAU;
  }
  
  void render(GameLoopHtml loop) {
    main.draw.ball(x, y, size * sizeMul.value(), phi + PI, speed);
  }
  
  void calculateSpeed() {
    /*
    if(speed < 1.5) {
      speed = speed * 1.15;
    } else {
      speed = speed + 0.25;
    }*/
    speed = sqrt(main.level/5.0) + 0.25;
  }

  void rebound() {
    
    sizeMul = new Line(1.4, 1.0, 0.1);
    
    var angleToCenter = atan2(x, y);
    var difference = angleToCenter - phi;
    var absDifference = difference.abs();

    if(absDifference > PI/2 && absDifference < (3 * PI/2)) {
      difference = difference + PI;
    }
    var deltaPhi = (main.deltaRotation * 15.0);
    var sign = sgn(deltaPhi);
    deltaPhi = log(deltaPhi.abs() + 1) * sign;

    // meh ...
    
    phi = phi + PI + (difference * 1) + ((main.rng.nextDouble() * REBOUND_RAND) - (REBOUND_RAND / 2)) - deltaPhi;
  }

}

class CellLayer implements GameObject {
  
  List<Cell> objects;
  
  int num;
  double targetRadius;
  Gen radiusFilter;
  double radius;
  Game main;
  
  
  bool active = true;
  
  
  CellLayer(this.main, this.num, this.targetRadius, this.radius, [int hitPoints=1]) {
    
    // vorischt, auch bei newRadius
    radiusFilter = new IIRFilter(new Line(this.radius, this.targetRadius, 0.2), 0.01, 0.99);
    
    objects = new List(num);
    
    for(int i=0; i < num; i++) {
      objects[i] = new Cell(main, this, i/num * TAU, THICKNESS, 1/num, hitPoints);
    }
  }
  
  void update(GameLoopHtml loop) {
    radius = radiusFilter.value();
    
    if(main.ball.distanceToOrigin + main.ball.size >= radius && 
       main.ball.distanceToOrigin - main.ball.size <= radius + THICKNESS) {
      int collisionIndex = (((main.ball.angleToOrigin - main.rotation) % TAU / TAU) * num).floor();
      Cell collisionObject = objects[collisionIndex];
      if(collisionObject.active && main.lastCollisionCell != collisionObject) {
        main.ball.rebound();
        main.score();
        main.lastCollisionCell = collisionObject;
        collisionObject.die();
        checkCells();
      }
    }
  }
  
  void demote() {
    for(var cell in objects) {
      cell.demote();
    }
  }
  
  void newRadius(double x) {
    targetRadius = x;
    radiusFilter = new IIRFilter(new Line(this.radius, x, 0.2), 0.01, 0.99);
  }
  
  void checkCells() {
    var alive = false;
    for(var cell in objects) {
      alive = alive || cell.active;
    }
    
    if(!alive) {
      die();
      main.repopulate();
    }
  }
  
  void die() {
    active = false;
    main.overRenders.add(new FlashCircle(main, radius, THICKNESS, 1.0));
  }
  
  void render(GameLoopHtml loop) {
    main.draw.blendAlpha();
    main.programs["basic"].bind();
    for(var object in objects) {
      object.render(loop);
    }
  }
}

class TimeCanvas {
  CanvasElement canvas;
  CanvasRenderingContext2D ctx;
  Game main;
  
  int dim = 512;
  
  String digit = "   ";
    
  GLTexture2D texture;
  
  TimeCanvas(this.main) {    
    canvas = new CanvasElement(width: dim, height: dim);
    ctx = canvas.context2D;
        
    texture = new GLTexture2D(main.gl, canvas, {"min": NEAREST, "mag": NEAREST});
  }
  
  void redraw() {
    ctx.clearRect(0, 0, dim, dim);
    
    ctx.fillStyle = "white";
    
    ctx.font = "${dim/4}pt 'Press Start 2P'";
    ctx.fillText(digit, 12, dim/1.48);
    
    texture.bind();
    texture.upload(canvas);
    texture.unbind();
  }
  
  void drawHelp() {
    ctx.clearRect(0, 0, dim, dim);
    
    ctx.fillStyle = "white";
    
    ctx.font = "${dim/24}pt 'Press Start 2P'";
    ctx.fillText("Click to Start", 12, dim/1.88 - 20);
    ctx.fillText("R to Restart", 12, dim/1.88 + 20);
    ctx.fillText("F for Fullscreen", 12, dim/1.88 + 60);

    // ctx.fillText("Press R to Restart", 12, dim/1.88 + 20);


    ctx.font = "${dim/16}pt 'Press Start 2P'";
    ctx.fillText("Ballkeeper", 12, dim/1.3 + 25);

    
    ctx.font = "${dim/32}pt 'Press Start 2P'";    
    ctx.fillText("Ludum Dare 28", 12, dim/1.3 + 60);
    ctx.fillText("@cappelnord", 12, dim/1.3 + 88);

    
    texture.bind();
    texture.upload(canvas);
    texture.unbind();    
  }
  
  void update(GameLoopHtml loop) {
    String newDigit = main.points.toString();
    if(newDigit.length < 3) {
      newDigit = "0${newDigit}";
    }
    if(newDigit.length < 3) {
      newDigit = "0${newDigit}";
    }
    
    if(digit != newDigit) {
      digit = newDigit;
      redraw();
    }
    
  }
}

class Background implements GameObject {
  
  Game main;
  
  List<Gen> xGens;
  List<Gen> yGens;
  List<Gen> sizeGens;
    
  List outerRadii = [1, 1.2, 1.4, 1.6, 1.8, 2];
  List<Gen> outerRotations;
  
  List<OverRenders> overRenders;
  
  final int numBlobs = 5;
  
  bool populated = false;
  
  Gen alpha;
  Gen jump;
  
  TimeCanvas timeCanvas;
  
  int darkCounter = 0;

  
  Background(this.main) {
    
    
    timeCanvas = new TimeCanvas(main);

    overRenders = new List();
    
    xGens = new List();
    yGens = new List();
    sizeGens = new List();
        
    for(int i = 0; i < numBlobs; i++) {
      xGens.add(new RndLines(9.0, 18.0, -1.1, 1.1));
      yGens.add(new RndLines(9.0, 18.0, -1.1, 1.1));
      sizeGens.add(new RndLines(8.0, 20.0, 2.3, 3.9));
    }
    
    alpha = new Const(0.0);
    jump = new Const(0.0);
    
    // document.body.children.add(timeCanvas.canvas);
    // document.body.children.add(new Element.html("<h3>TEST</h3>"));
  }
  
  void populate() {
    outerRotations = new List(outerRadii.length);
    for(int i = 0; i < outerRadii.length; i++) {
      outerRotations[i] = new IIRFilter(new RndLines(20.0, 30.0, -TAU, TAU), 0.05, 0.95);
    }
    
    alpha = new Line(0.0, 1.0, 6.0);
    populated = true;
  }
  
  void update(GameLoopHtml loop) {
    if(main.rng.nextDouble() > 0.95 && darkCounter > 400) {
      // overRenders.add(new DarkCircle(main));
      darkCounter = 0;
    }
    darkCounter++;
  }
  
  void render(GameLoopHtml loop) {
    
    timeCanvas.update(loop);
        
    main.draw.blendAdd();
    if(!main.lost) {
      main.draw.color = [0.02, 0.12, 0.2, 1.0];
    } else {
      main.draw.color = [0.17, 0.06, 0.02, 1.0];
    }
    for(int i = 0; i < numBlobs; i++) {
      var dim = sizeGens[i].value();
      main.draw.texQuadC(main.textures["blur"], xGens[i].value(), yGens[i].value(), dim, dim);
    }
    
    if(main.started && !main.lost) {
      main.draw.blendSub();
      main.draw.color = [1.0, 1.0, 1.0, alpha.value() * 0.4];
      var size = 0.15 + jump.value() * 0.035;
      main.draw.texQuadC(timeCanvas.texture, 0, 0, size, size);
    }
    
    
    if(populated) {
      
      main.draw.blendSub();
      overRenders.retainWhere((x) => !x.finished());
      for(var x in overRenders) {
        x.render();
      }
      
      GLProgram prog = main.programs["basic"];
      prog.bind();
      prog.uniforms["color"].set([1.0, 1.0, 1.0, 0.09 * alpha.value()]);
      main.draw.blendAdd();
      
      for(int i = 0; i < outerRadii.length; i++) {
        var radius = outerRadii[i];
        var rotation = outerRotations[i].value();
        main.draw.arc(prog, rotation, radius, THICKNESS, 0.5);
      }
    }
  }
}

class Cell implements GameObject {
  Game main;
  CellLayer layer;
  num phi;
  num thickness;
  num length;
  
  Gen alpha;
  Gen hitTransition;
  
  bool active = true;
  bool doDraw = true;
  
  int hitPoints;
  
  Cell(this.main, this.layer, this.phi, this.thickness, this.length, [this.hitPoints=1]) {
    alpha = new Line(0.0, 1.0, 1.0);
    if(hitPoints == 2) {
      hitTransition =  new Const(1.0);
    } else {
      hitTransition = new Const(0.0);
    }
  }
  
  void update(GameLoopHtml loop) {
    if(active) {
      
    }
  }
  
  void demote() {
    if(hitPoints > 1) {
      die(1.0);
    }
  }
  
  void die([double fade=0.0]) {
    hitPoints--;
    if(hitPoints == 1) {
      hitTransition = new Line(1.0, 0.0, fade);
    }
    if(hitPoints <= 0) {
      active = false;
      alpha = new Line(1.0, 0.0, 0.25);
    }
  }
  
  void render(GameLoopHtml loop) {
    if(doDraw) {
      var renderLength = length * (1 - CELL_SPACING);
      
      main.programs["basic"].uniforms["color"].set([1.0, hitTransition.value() * 0.5 , 1.0 - hitTransition.value() , alpha.value()]);
      main.draw.arc(main.programs["basic"], main.rotation + phi, layer.radius, thickness, renderLength);
    }
  }
}

class GameDrawHelper {
  RenderingContext gl;
  Game main;
  Map<String, GLProgram> programs;
  
  Buffer arcBuffer;
  int arcBufferSegments = 64;
  
  Buffer ballBuffer;
  int ballBufferSegments = 64;
  
  Buffer texVertBuffer;
  Buffer texUVBuffer;
  
  int width, height;
  double ratio;
  
  List color = [1.0, 1.0, 1.0, 1.0];

  GameDrawHelper(this.main, this.programs) {
    gl = main.gl;
    generateArcBuffer();
    generateBallBuffer();
    generateTexBuffer();
  }
  
  void blendAdd() {
    gl.blendEquation(FUNC_ADD);
    gl.blendFunc(SRC_ALPHA, ONE);
  }
  
  void blendAlpha() {
    gl.blendEquation(FUNC_ADD);
    gl.blendFunc(SRC_ALPHA, ONE_MINUS_SRC_ALPHA);
  }
  
  void blendSub() {
    gl.blendEquation(FUNC_REVERSE_SUBTRACT);
    gl.blendFunc(SRC_ALPHA, ONE);
  }
  
  void texQuadC(GLTexture2D tex, num x, num y, num dimX, num dimY, [num rotation=0]) {
    main.programs["tex"].bind();
    tex.activate(0);
    main.programs["tex"].uniforms["tex"].set(0);
    main.programs["tex"].uniforms["color"].set(color);
    quadC(main.programs["tex"], x, y, dimX, dimY, rotation);
  }
  
  void quadC(GLProgram program, num x, num y, num dimX, num dimY, [num rotation=0]) {
    program.uniforms["pos"].set([x, y]);
    program.uniforms["dim"].set([dimX, dimY]);
    program.uniforms["rotation"].set(rotation);
    program.uniforms["ratio"].set(ratio);
    
    program.attributes["verts"].set(texVertBuffer, 2);
    program.attributes["uvs"].set(texUVBuffer, 2);
    gl.drawArrays(TRIANGLES, 0, 6);
  }
  
  void arc(GLProgram program, num rotation, num radius, num thickness, num length) {

    program.uniforms["rotation"].set(rotation);
    program.uniforms["radius"].set(radius);
    program.uniforms["thickness"].set(thickness);
    program.uniforms["length"].set(length);
    program.uniforms["ratio"].set(ratio);
    
    program.attributes["verts"].set(arcBuffer, 2);
    gl.drawArrays(TRIANGLE_STRIP, 0, arcBufferSegments * 2);
  }
  
  void ball(num x, num y, num size, double rotation, double speed) {
    GLProgram program = main.programs["ball"];
    program.bind();
    program.uniforms["pos"].set([x, y]);
    program.uniforms["size"].set(size);
    program.uniforms["ratio"].set(ratio);
    program.uniforms["rotation"].set(rotation);
    program.uniforms["speed"].set(0);
    
    program.attributes["verts"].set(ballBuffer, 2);
    gl.drawArrays(TRIANGLE_FAN, 0, ballBufferSegments+1);
  }
  
  void generateArcBuffer() {
    Float32List arr = new Float32List(arcBufferSegments * 4);
    for(int i = 0; i < arr.length; i = i + 4) {
      num x = i / 4 / arcBufferSegments;
      arr[i] = x;
      arr[i+1] = 0.0;
      arr[i+2] = x;
      arr[i+3] = 1.0;
    }
    
    arcBuffer = gl.createBuffer();
    gl.bindBuffer(ARRAY_BUFFER, arcBuffer);
    gl.bufferData(ARRAY_BUFFER, arr, STATIC_DRAW);
  }
  
  void generateBallBuffer() {
    Float32List arr = new Float32List(ballBufferSegments * 2 + 2);
    arr[0] = 0.0;
    arr[1] = 0.0;
    for(int i = 0; i < ballBufferSegments; i = i + 2) {
      arr[i+2] = sin(i/(ballBufferSegments-2) * TAU);
      arr[i+3] = cos(i/(ballBufferSegments-2) * TAU);
    }
    
    ballBuffer = gl.createBuffer();
    gl.bindBuffer(ARRAY_BUFFER, ballBuffer);
    gl.bufferData(ARRAY_BUFFER, arr, STATIC_DRAW);
  }
  
  void generateTexBuffer() {
    Float32List vArr = new Float32List.fromList([
       -1.0,  1.0, -1.0, -1.0,  1.0, 1.0,
       -1.0, -1.0,  1.0, -1.0,  1.0, 1.0
    ]);
    
    Float32List tArr = new Float32List.fromList([
       0.0, 0.0, 0.0, 1.0, 1.0, 0.0,
       0.0, 1.0, 1.0, 1.0, 1.0, 0.0
    ]);
    
    texVertBuffer = gl.createBuffer();
    gl.bindBuffer(ARRAY_BUFFER, texVertBuffer);
    gl.bufferData(ARRAY_BUFFER, vArr, STATIC_DRAW);
    
    texUVBuffer = gl.createBuffer();
    gl.bindBuffer(ARRAY_BUFFER, texUVBuffer);
    gl.bufferData(ARRAY_BUFFER, tArr, STATIC_DRAW);
  }
}